@@ -3,6 +3,11 @@ import os
 from distutils.version import LooseVersion
 import pkg_resources
 from mlagents.torch_utils import cpu_utils
+from mlagents.trainers.settings import TorchSettings
+from mlagents_envs.logging_util import get_logger
+
+
+logger = get_logger(__name__)
 
 
 def assert_torch_installed():
@@ -32,16 +37,32 @@ import torch  # noqa I201
 torch.set_num_threads(cpu_utils.get_num_threads_to_use())
 os.environ["KMP_BLOCKTIME"] = "0"
 
-# Known PyLint compatibility with PyTorch https://github.com/pytorch/pytorch/issues/701
-# pylint: disable=E1101
-if torch.cuda.is_available():
-    torch.set_default_tensor_type(torch.cuda.FloatTensor)
-    device = torch.device("cuda")
-else:
-    torch.set_default_tensor_type(torch.FloatTensor)
-    device = torch.device("cpu")
+
+_device = torch.device("cpu")
+
+
+def set_torch_config(torch_settings: TorchSettings) -> None:
+    global _device
+
+    if torch_settings.device is None:
+        device_str = "cuda" if torch.cuda.is_available() else "cpu"
+    else:
+        device_str = torch_settings.device
+
+    _device = torch.device(device_str)
+
+    if _device.type == "cuda":
+        torch.set_default_tensor_type(torch.cuda.FloatTensor)
+    else:
+        torch.set_default_tensor_type(torch.FloatTensor)
+    logger.info(f"default Torch device: {_device}")
+
+
+# Initialize to default settings
+set_torch_config(TorchSettings(device=None))
+
 nn = torch.nn
 
 
 def default_device():
-    return device
+    return _device
